"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
import uuid, sys, os.path
from node import Node
from libs import base
from constants import *
from datetime import datetime
from utilities import connection_manager
from helpers.seed_down_handler import HandleSeedDown
from helpers.seed_down_handler_with_version import HandleSeedDownWithVersion
from helpers import ip_checker
from operator import itemgetter

ERROR_DESCRIPTION = {1: "Unknown Server Error",
     50: "Last 'proto' in a Chunked Sequence",
     51: "Security Not Supported",
     52: "Security Not Enabled",
     53: "Security Scheme Not Supported",
     54: "Unrecognized Command ID",
     55: "Unrecognized Field ID",
     56: "Illegal State (e.g. valid but unexpected command)",
     60: "No User or Unrecognized User",
     61: "User Already Exists",
     62: "No Password or Bad Password",
     65: "No Credential or Bad Credential",
     70: "No Role(s) or Unrecognized Role(s)",
     80: "Not Authenticated",
     81: "Role/Privilege Violation"}

class Cluster(object):
    """Cluster contains all information related to cluster"""

    def __new__(cls, ip, port):

        if base.is_node_available(ip, port) is False:
            return UNABLE_TO_CONNECT
        return super(Cluster, cls).__new__(cls, ip, port)

    def __init__(self, ip, port):

        self._ip = ip
        self._port = port
        self._seed = "%s:%s" % (ip, port)
        self._cluster_id = str(uuid.uuid4())

        self._nodes = {self._seed: Node(self._seed, self._cluster_id)}
        self._node_id = self._nodes[self._seed].get_node_id()

        self._node_list = {}
        if self._node_id is not None:
            self._node_list = {self._node_id :self._seed}

        self._last_access_time = datetime.now()
        self._id = None
        self._current_nodes = []
        self._index_list = {}
        self._node_visibilties = {}

        self.get_updates()
        self.set_status_on()
        self._latest_build_no = self.get_build_no_of_cluster_nodes().get("latest_build_no")
        sys.stderr.write("Aerospike Version for cluster with seed node %s: %s\n" % (self._ip, str(self._latest_build_no)))
        self._namespaces = self.get_namespaces()
        self._cluster_key = self._nodes[self._seed].get_cluster_key()

    def get_cluster_id(self):

        return self._cluster_id

    def get_updates(self, alerts_flag=False):

        enterprise_flag = os.path.isfile(ENTERPRISE_FLAG_FILE)
        node_list = {}
        if enterprise_flag and alerts_flag:
            cluster_integrity = "true"
            is_cluster_size_same = True
            node_cluster_size = None

        for node in list(self._nodes.values()):
            if not node.is_valid():
                continue
            node_list.update(node.get_node_list())

            if enterprise_flag and alerts_flag:
                if node.get_statistic("cluster_integrity") == "false":
                    cluster_integrity = "false"
                current_node_cluster_size = node.get_statistic("cluster_size")
                if current_node_cluster_size != NOT_SUPPORTED:
                    if node_cluster_size is None:
                        node_cluster_size = current_node_cluster_size
                    elif node_cluster_size != current_node_cluster_size:
                        is_cluster_size_same = False

        if enterprise_flag and alerts_flag:
            self._alert.on_cluster_integrity_change(cluster_integrity)
            self._alert.on_cluster_size_change(is_cluster_size_same)

        if node_list is not None:
            new_node_list = {}
            for key in node_list:
                if key not in self._node_list:
                    new_node_list[key] = node_list[key]
            for new_node_id, new_node_address in new_node_list.items():
                for node_id, node_address in self._node_list.items():
                    if node_address == new_node_address:
                        del self._node_list[node_id]
                self._node_list.update({new_node_id: new_node_address})
                if new_node_address not in self._nodes:
                    self._nodes[new_node_address] = Node(new_node_address,
                        self._cluster_id)
                if enterprise_flag and alerts_flag:
                    self._alert.on_new_node_add(new_node_address)
        self.calculate_visibilities()

    def add_existing_node(self, node_address):

        ip, port = node_address.split(":")
        ip = ip_checker.check_ip(ip)
        node_id = base.get_node_id(ip, port)

        if ip == "Unable to connect":
            return dict(status="failure", error="Unable to connect")
        elif ip == "Unable to resolve hostname":
            return dict(status="failure", error="Unable to resolve hostname")
        elif ip == "Please enter network IP":
            return dict(status="failure", error="Please enter network IP")
        else:
            if (node_id is not None
                and node_id in self._node_list.keys()):
                return dict(status="failure", error="Node already exists")
            else:
                new_node_obj = Node(node_address, self._cluster_id,
                    validity=None)
                if (new_node_obj.get_status() == "on" and
                    self.get_cluster_key() != new_node_obj.get_cluster_key()):
                    message = "Node <strong>%s</strong>" % node_address
                    message += " cannot be monitored here as it belongs"
                    message += " to a different cluster."
                    return dict(status=message)
                else:
                    self._nodes[node_address] = new_node_obj
                    if node_id is None:
                        node_id = node_address
                    self._node_list.update({node_id: node_address})
                    return dict(status="success")

    """
    def add_new_node(self, node_address):

        ip, port = node_address.split(":")
        node_address = base.get_self_address(ip, port)
        if node_address == UNABLE_TO_CONNECT or node_address in self._node_list:
            return None
        self._node_list.add(node_address)
        self._nodes[node_address] = Node(node_address, self._cluster_id)
        return node_address
    """

    def remove_node(self, node_address):
        for key, value in self._node_list.items():
            if node_address == value and node_address in self._nodes:
                del self._node_list[key]
                del self._nodes[node_address]
                return dict(status="success")
        return dict(status="failure")

    def get_id(self):
        return self._id

    def get_seed_ip(self):
        return self._ip

    def get_seed_port(self):
        return self._port

    def get_seed(self):
        return self._seed

    def set_seed(self, seed):
        self._seed = seed
        self._ip, self._port = seed.split(":")

    def get_status(self):
        return self._status

    def set_status_off(self):
        self._status = "off"

    def set_status_on(self):
        self._status = "on"

    def get_last_access_time(self):
        return self._last_access_time

    def update_last_access_time(self):
        self._last_access_time = datetime.now()

    def get_nodes(self):

        return self._nodes

    @HandleSeedDown()
    def get_build_no(self, ip=None, port=None):

        ip = ip or self._ip
        port = port or self._port

        address = "%s:%s" % (self._ip, self._port)
        node_obj = self.get_node(address)

        if node_obj is not None:
            sock = node_obj.get_socket_from_queue()
            build_no = base.get_node_build(self._ip, self._port, sock=sock)
        else:
            build_no = base.get_node_build(self._ip, self._port)
        if build_no is None:
            new_sock = connection_manager.create_socket(self._ip, self._port)
            if new_sock is not None:
                sock = new_sock
                build_no = base.get_node_build(self._ip, self._port, sock=sock)
        if node_obj is not None:
            node_obj.put_socket_in_queue(sock)

        return build_no

    def get_build_no_of_cluster_nodes(self):

        info = dict(version_list=dict())
        for node in list(self._nodes.values()):
            build_no = node.get_build()
            if build_no is not None:
                if build_no in list(info['version_list'].keys()):
                    info['version_list'][build_no].append(node.get_address())
                else:
                    info['version_list'][build_no] = [node.get_address()]
        build_list = []
        for build_no in list(info['version_list'].keys()):
            build_list.append(build_no.replace('.', ''))
        if list(info['version_list'].keys()) == []:
            info['latest_build_no'] = "Undefined"
        else:
            info['latest_build_no'] = list(info['version_list'].keys())[build_list.index(max(build_list))]
        return info

    def get_latest_build_no(self):

        return self._latest_build_no

    def get_node_list(self):

        return self._node_list

    def get_off_nodes(self):

        off_nodes = []
        for node in list(self._nodes.values()):
            if node.get_status() == "off":
                off_nodes.append(node.get_address())
        return off_nodes

    def get_different_cluster_nodes(self):

        different_cluster_nodes = []
        for node in list(self._nodes.values()):
            if node.is_valid() is False:
                different_cluster_nodes.append(node.get_address())
                self.remove_node(node.get_address())
        return different_cluster_nodes

    def get_number_of_nodes(self):

        return len(self._node_list)

    def get_namespaces(self):

        self._namespaces = self.update_namespaces()
        return self._namespaces

    def get_number_of_namespaces(self):

        return len(self.get_namespaces())

    def update_namespaces(self):

        namespace_list = set()
        for node in list(self._nodes.values()):
            namespace_list = namespace_list.union(node.get_namespace_list())
        return list(namespace_list)

    def get_memory_usage(self):

        total_bytes_memory = 0
        used_bytes_memory = 0
        details = dict()

        for node_obj in list(self._nodes.values()):
            if node_obj is None or not node_obj.is_valid():
                continue

            used, total = node_obj.get_memory()
            if used.isdigit() and total.isdigit():
                total_bytes_memory += int(total)
                used_bytes_memory += int(used)
                details[node_obj.get_address()] = dict(used=int(used), free=int(total)-int(used))

        total_free_memory = total_bytes_memory - used_bytes_memory

        return dict(used=used_bytes_memory, free=total_free_memory, details=details)

    def get_disk_usage(self):

        total_bytes_disk = 0
        used_bytes_disk = 0
        details = dict()

        for node_obj in list(self._nodes.values()):
            if node_obj is None or not node_obj.is_valid():
                continue

            used, total = node_obj.get_disk()
            if used.isdigit() and total.isdigit():
                total_bytes_disk += int(total)
                used_bytes_disk += int(used)
                details[node_obj.get_address()] = dict(used=int(used), free=int(total)-int(used))

        total_free_disk = total_bytes_disk - used_bytes_disk

        return dict(used=used_bytes_disk, free=total_free_disk, details=details)

    def get_throughput(self, node_addresses):

        throughput = dict()
        for node_address in node_addresses:
            node = self.get_node(node_address)
            if node is None or node.get_status() == "off" or not node_obj.is_valid():
                continue
            throughput[node_address] = node.get_throughput()
        return throughput

    def get_statistics(self):

        self._statistics = []
        for node in list(self._nodes.values()):
            if node.is_valid():
                stat = dict(address=node.get_address())
                stat.update(node.get_updated_statistics(self))
                self._statistics.append(stat)
        return self._statistics

    def update_statistics(self):

        for node in list(self._nodes.values()):
            if node.is_valid() is not False:
                node.get_updated_statistics(self)

    def calculate_visibilities(self):

        for node_obj in list(self._nodes.values()):
            if not node_obj.is_valid():
                continue
            nodes = node_obj.get_node_list().values()

            node_status = False if nodes == set() else True

            old_visibility = node_obj.get_cluster_visibility()
            visibility = True

            for node2_obj in list(self._nodes.values()):
                if not node2_obj.is_valid():
                    continue
                nodes2 = node2_obj.get_node_list().values()
                if node_obj.get_address() == node2_obj.get_address() or nodes2 == set():
                    continue

                if node_obj.get_address() in nodes2:
                    if nodes != set():
                        visibility &= True
                    else:
                        node_status = True
                elif node_obj.get_address() not in nodes2:
                    if nodes == set():
                        node_status |= False
                    else:
                        visibility = False
                        break

            if old_visibility != visibility:
                node_obj.set_cluster_visibility(visibility)
            if node_status == False:
                node_obj.set_status("off")
            else:
                node_obj.set_status("on")

    def get_node_visibility(self, node_address):

        return self._node_visibilties.get(node_address, None)

    def get_node(self, node_address):

        return self._nodes.get(node_address, None)

    def get_namespace(self, namespace_name, node_address):

        node_obj = self.get_node(node_address)
        if node_obj is None or not node_obj.is_valid():
            return None
        return node_obj.get_namespace(namespace_name)

    def get_xdr(self, xdr_port, node_address):

        node_obj = self.get_node(node_address)
        if node_obj is None or not node_obj.is_valid():
            return None
        return node_obj.get_xdr(xdr_port)

    def get_dc_names(self, xdr_port="3004"):

        dc_names = dict()
        for node in list(self._nodes.values()):
            address = node.get_address()
            xdr = self.get_xdr(xdr_port, address)
            if xdr is None:
                dc_names[address] = "node off"
            elif xdr == {}:
                dc_names[address] = "xdr off"
            else:
                dc_names[address] = xdr.get_dc_names()
        return dc_names

    @HandleSeedDownWithVersion()
    def get_indexes(self, ip=None, port=None):

        ip = ip or self._ip
        port = port or self._port
        address = "%s:%s" % (ip, port)
        node_obj = self.get_node(address)

        try:
            if node_obj is not None:
                sock = node_obj.get_socket_from_queue()
                result = base.get_indexes(ip, port, sock=sock)
            else:
                result = base.get_indexes(ip, port)
            if type(result) is int:
                new_sock = connection_manager.create_socket(ip, port)
                if new_sock is not None:
                    sock=new_sock
                    result = base.get_indexes(ip, port, sock=sock)
        except:
            result = UNSUPPORTED_VERSION
        node_obj.put_socket_in_queue(sock)
        return result

    def get_indexes_by_namespaces(self, namespace):

        indexes = self.get_indexes()
        if indexes == UNABLE_TO_CONNECT:
            return UNABLE_TO_CONNECT
        elif indexes == UNSUPPORTED_VERSION:
            indexes = {}

        return indexes.get(namespace, [])

    def get_index_list_by_namespace(self):

        index_details = self.get_indexes()
        if index_details == UNABLE_TO_CONNECT:
            return {}
        elif index_details == UNSUPPORTED_VERSION:
            index_details = {}

        index_list = {}
        for namespace, indexes in list(index_details.items()):
            index_list[namespace] = [index["indexname"] for index in indexes]
        return index_list

    @HandleSeedDown()
    def get_sets_info(self, ip=None, port=None):

        first_iteration = True
        previous_sets_info = None
        for node_obj in list(self._nodes.values()):
            address = node_obj.get_address()
            ip, port = str(address).split(':')
            sock = node_obj.get_socket_from_queue()
            sets_info = base.get_sets_info(ip, port, sock=sock)
            if type(sets_info) is int:
                new_sock = connection_manager.create_socket(ip, port)
                if new_sock is not None:
                    sock = new_sock
                    sets_info = base.get_sets_info(ip, port, sock=sock)
            node_obj.put_socket_in_queue(sock)
            if sets_info != {} and sets_info != UNABLE_TO_CONNECT:
                if first_iteration is False:
                    for key in sets_info.iterkeys():
                        sets = sets_info[key]
                        for set_obj in sets:
                            previous_sets = previous_sets_info[key]
                            try:
                                set_index = map(itemgetter('set_name'),previous_sets).index(set_obj['set_name'])
                            except ValueError:
                                previous_sets.append(set_obj)
                                continue
                            #set_index = previous_sets.index(set_obj)
                            previous_set = previous_sets[set_index]
                            previous_set['n_objects'] =  int(previous_set['n_objects']) + int(set_obj['n_objects'])
                else:
                    first_iteration = False
                    previous_sets_info = sets_info
            #else:
                #previous_sets_info = sets_info
        return previous_sets_info

    def get_sets_info_by_namespaces(self, namespace):

        sets_info = self.get_sets_info()
        if sets_info == UNABLE_TO_CONNECT:
            return UNABLE_TO_CONNECT
        if sets_info is None:
            return {}
        else:
            return sets_info.get(namespace, [])

    @HandleSeedDown()
    def get_cluster_key(self, ip=None, port=None):

        try:
            self._cluster_key = self._nodes[self._seed].get_cluster_key()
        except KeyError:
            return None
        return self._cluster_key

    def close_sockets(self):

        for node in list(self._nodes.values()):
            node.close_sockets()


