"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
import sys
import functools
import datetime
import threading
import traceback
from libs import base
from flask import json, current_app
from models import cluster
from constants import UNABLE_TO_CONNECT, CLUSTER_TIMEOUT, ERROR_OCCURRED, EXISTING_CLUSTER
from helpers.synchronizing import synchronous


class ClusterManager(object):
    """It manages the all current cluster objects"""

    def __init__(self):

        self._clusters = {}
        self.clusters_lock = threading.RLock()

    @synchronous('clusters_lock')
    def create_cluster(self, node_address):

        cluster_info = self.get_cluster_by_node(node_address)
        if cluster_info is None:
            ip, port = node_address.split(":")
            return cluster.Cluster(ip, port)
        return EXISTING_CLUSTER

    def add_cluster(self, cluster):

        self._clusters[cluster.get_cluster_id()] = cluster

    def remove_cluster(self, cluster_id):

        del(self._clusters[cluster_id])

    def get_cluster_by_id(self, cluster_id):

        return self._clusters.get(cluster_id, None)

    def get_cluster_by_node(self, node_address):

        for cluster_id, cluster_obj in list(self._clusters.items()):
            if node_address in cluster_obj.get_nodes():
                return dict(cluster_id=cluster_id, nodes=cluster_obj.get_node_list())
        return None

    def get_cluster_info(self, host_address):

        update_interval = current_app.config.get("UPDATE_INTERVAL")
        try:
            ip, port = host_address.split(":")
        except:
            return dict(error="Unable to connect - %s" % host_address)

        node_address = base.get_self_address(ip, port)
        if node_address == UNABLE_TO_CONNECT:
            return dict(error="Unable to connect - %s" % host_address)

        cluster_info = self.get_cluster_by_node(node_address)
        if cluster_info is not None:
            cluster_info.update(update_interval=update_interval)
            return cluster_info

        cluster = self.create_cluster(node_address)
        if cluster == UNABLE_TO_CONNECT or cluster == ERROR_OCCURRED:
            return dict(error="Unable to connect - %s" % host_address)
        elif cluster == EXISTING_CLUSTER:
            cluster_info = self.get_cluster_by_node(node_address)
            if cluster_info is not None:
                cluster_info.update(update_interval=update_interval)
                return cluster_info
        else:
            self.add_cluster(cluster)
            return dict(cluster_id=cluster.get_cluster_id(), nodes=cluster.get_node_list(), update_interval=update_interval)

    def get_clusters(self):

        return self._clusters

    def remove_unused_clusters(self):

        current_time = datetime.datetime.now()

        for cluster_id, cluster_obj in list(self._clusters.items()):
            elapsed_time = current_time - cluster_obj.get_last_access_time()
            if elapsed_time.seconds > CLUSTER_TIMEOUT * 60:
                self.remove_cluster(cluster_id)

    def check_cluster_validity(self):

        def decorator(fn):

            @functools.wraps(fn)
            def f(*args, **kwargs):

                cluster = self.get_cluster_by_id(kwargs["cluster"])
                if cluster is None:
                    return json.dumps(dict(error="Invalid cluster id : %s" % kwargs["cluster"]))

                cluster.update_last_access_time()
                kwargs["cluster"] = cluster
                try:
                    return fn(*args, **kwargs)
                except Exception:
                    sys.stderr.write("args:%s\nkwargs:%s" % (args, kwargs))
                    sys.stderr.write((traceback.format_exc()))
                    raise Exception("Internal server error!")

            return f
        return decorator
